package example

import AuthService._
import example.pocketbase.Api
import upickle.default._
import example.pocketbase.Models._
import java.time.Instant
import org.thymeleaf.TemplateEngine
import org.thymeleaf.templateresolver.ClassLoaderTemplateResolver
import org.thymeleaf.context.{Context => PageContext}
import scala.jdk.CollectionConverters._

case class AuthService()(implicit cc: castor.Context, log: cask.Logger)
    extends cask.Routes {

  @cask.get("/")
  def getIndex(request: cask.Request) = {
    val authCookieOpt = request.cookies.get(authCookieName)
    println(s"getting index with auth cookie: $authCookieOpt")

    authCookieOpt match {
      case None =>
        println("cookie was None")
        val redirectToLogin = renderRedirectPage(
          "/login",
          "You will be redirected to login page soon",
          0
        )
        cask.Response(
          redirectToLogin,
          200,
          Seq(
            "Content-Type" -> "text/html;charset=UTF-8"
          )
        )
      case Some(authCookie) =>
        val jwt = authCookie.value
        val refreshedSession = pocketbaseApi.refreshSession(jwt)
        refreshedSession match {
          case Right(freshAuth) =>
            val context = new PageContext()
            context.setVariable("userRecord", freshAuth.record)

            val indexPage = templateEngine.process("index", context)
            cask.Response(
              indexPage,
              headers = Seq("Content-Type" -> "text/html;charset=UTF-8"),
              cookies =
                Seq(cask.Cookie(name = authCookieName, value = freshAuth.token))
            )
          case Left(err) =>
            val indexHtml = """
<h1>Ok, not good at all</h1>
<p>For some reason error during authorization</p>
<p>You can try again <a href="/login">here</a>
"""
            cask.Response(
              indexHtml,
              headers = Seq("Content-Type" -> "text/html;charset=UTF-8"),
              cookies = Seq(
                cask.Cookie(
                  name = authCookieName,
                  value = "",
                  expires = Instant.EPOCH
                )
              )
            )
        }
    }
  }

  @cask.get("/login")
  def getLoginPage() = {
    // render auth page with the available oauth providers
    println("landed on login page")
    val authOptions = pocketbaseApi.listAuthMethods()

    val iconLocations = Map(
      "github" -> "public/icons/github-svgrepo-com(1).svg",
      "apple" -> "public/icons/apple-svgrepo-com.svg",
      "google" -> "public/icons/google-svgrepo-com.svg",
      "facebook" -> "public/icons/facebook-f-svgrepo-com.svg",
      "gitlab" -> "public/icons/gitlab-svgrepo-com(1).svg"
    )

    val oauthLinks = authOptions.authProviders
      .map(provider => provider.name -> getRedirectLandingUrl(provider.name))
      .toMap

    val context = new PageContext()
    context.setVariable("iconLocationsMap", iconLocations.asJava)
    context.setVariable(
      "oauthProvidersList",
      authOptions.authProviders.sortBy(_.name).asJava
    )
    context.setVariable("redirectUrl", oauthLinks.asJava)

    val authPage = templateEngine.process("auth", context)

    val oauthVerificationCookie = authOptions.toOauthCookieInfo()
    val oauthCookieEncoded = write(oauthVerificationCookie)

    val githubOption = authOptions.authProviders.find(_.name == "github")

    val githubRedirect =
      githubOption.map(_.authUrl).getOrElse("") ++ getRedirectLandingUrl(
        "github"
      )

    cask.Response(
      authPage,
      headers = Seq("Content-Type" -> "text/html;charset=UTF-8"),
      cookies = Seq(
        cask.Cookie(
          name = oauthVerifiersCookieName,
          value = oauthCookieEncoded,
          path = "/"
        )
      )
    )

  }

  /** @oauthVerifiers
    *   \- required cookie from start of oauth initialization cask will fail
    *   request if this is not present
    */
  @cask.get(s"${baseRedirectUrl}/:provider")
  def receiveOauthRedirect(
      provider: String,
      state: String,
      code: String,
      scope: Option[String] = None,
      prompt: Option[String] = None,
      authuser: Option[String] = None,
      request: cask.Request
  ) = {
    println(
      s"received redirect for $provider with state: $state and code: $code"
    )

    val authVerifierOpt = request.cookies
      .get(oauthVerifiersCookieName)
      .map(cookie => read[OauthInfoCookie](cookie.value))
      .flatMap(_.providersVerification.find(_.name == provider))

    val resultOpt = for {
      authVerifier <- authVerifierOpt
      _ <- Some(()).filter(_ => state == authVerifier.state)
      pocketbaseAuthResult <- pocketbaseApi
        .authWithOauth(
          provider = provider,
          code = code,
          verifier = authVerifier.codeVerifier,
          redirectUrl = getRedirectLandingUrl(provider)
        )
        .toOption
    } yield pocketbaseAuthResult

    val okMessageFirst = resultOpt match {
      case None =>
        // i guess with the SSR i'll need to return message about unsuccessful auth?
        val unsuccessHtml = s"""
<h1>Auth unsuccessful</h1>
"""
        cask.Response(
          unsuccessHtml,
          headers = Seq("Content-Type" -> "text/html;charset=UTF-8")
        )
      case Some(result) =>
        // this is already fully successful auth
        val redirectToIndex = renderRedirectPage(
          "/",
          "You will be redirected to home page soon",
          0
        )

        cask.Response(
          redirectToIndex,
          headers = Seq("Content-Type" -> "text/html;charset=UTF-8"),
          cookies = Seq(
            cask.Cookie(
              name = authCookieName,
              value = result.token,
              path = "/"
            ),
            cask.Cookie(
              name = oauthVerifiersCookieName,
              value = "",
              expires = Instant.EPOCH,
              path = "/"
            )
          )
        )
    }
    okMessageFirst
  }

  @cask.get(s"/logout")
  def logoutAndRedirectToRoot() = {
    println(s"about to logout")

    val redirectToIndex = renderRedirectPage(
      "/",
      "You will be redirected to home page soon",
      0
    )

    cask.Response(
      redirectToIndex,
      headers = Seq("Content-Type" -> "text/html;charset=UTF-8"),
      cookies = Seq(
        cask.Cookie(
          name = authCookieName,
          value = "",
          path = "/",
          expires = Instant.EPOCH
        )
      )
    )
  }

  @cask.staticResources("/public")
  def servePublicResources() = "public"

  initialize()
}

object AuthService {
  val authCookieName = "auth"
  // this is to share info between /auth page and /redirect-landing
  // can be improved by generating same state and codeVerifier for all oauth links
  // and can also be stored purely on the backend - a tad more secure, i guess
  // but if cookie is under https, should be ok
  val oauthVerifiersCookieName = "oauthVerifiers"

  val templateEngine = {
    val templateResolver = new ClassLoaderTemplateResolver()
    templateResolver.setPrefix("templates/");
    templateResolver.setSuffix(".html")
    templateResolver.setTemplateMode("HTML5")

    val templateEngine = new TemplateEngine()
    templateEngine.setTemplateResolver(templateResolver)
    templateEngine
  }

  val pocketbaseApi = Api("http://127.0.0.1:8090")
  val selfUri = "http://127.0.0.1:8080"

  // i guess TOOD put that into config
  val baseRedirectUrl = "/api/oauth2-redirect"

  /*
   * let's have separate redirect urls for different providers
   * if we'd have same url for all providers, we'll need way to figure out from state which one was called on front end.
   * it's possible, but path param looks simpler right now
   * github redirect : http://127.0.0.1:8080/api/oauth2-redirect/github
   */
  def getRedirectLandingUrl(provider: String): String =
    s"${selfUri}${baseRedirectUrl}/${provider}"

  /*
   * this is a page of my site that does redirecting
   * from index to register
   * or from oauth provider redirect landing to index
   */
  def renderRedirectPage(
      redirectTo: String,
      message: String,
      redirectAfter: Int = 0
  ): String = {
    val context = new PageContext()
    context.setVariable("redirectTo", redirectTo)
    context.setVariable("message", message)
    context.setVariable("redirectAfter", redirectAfter)
    templateEngine.process("redirect", context)
  }

}
